# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import logging
import threading
import datetime
import warnings
from typing import TYPE_CHECKING, List, Literal, Union, Any, Callable, Optional, Dict, Tuple, overload

from ._client_base import ClientBase
from ._consumer import EventHubConsumer
from ._constants import ALL_PARTITIONS, TransportType
from ._eventprocessor.event_processor import EventProcessor
from ._eventprocessor.common import LoadBalancingStrategy


if TYPE_CHECKING:
    from ssl import SSLContext
    from ._eventprocessor.partition_context import PartitionContext
    from ._common import EventData
    from ._client_base import CredentialTypes
    from ._eventprocessor.common import CloseReason
    from ._eventprocessor.checkpoint_store import CheckpointStore

_LOGGER = logging.getLogger(__name__)


class EventHubConsumerClient(ClientBase):  # pylint: disable=client-accepts-api-version-keyword

    # pylint: disable=client-method-missing-tracing-decorator
    """The EventHubConsumerClient class defines a high level interface for
    receiving events from the Azure Event Hubs service.

    The main goal of `EventHubConsumerClient` is to receive events from all partitions of an EventHub with
    load-balancing and checkpointing.

    When multiple `EventHubConsumerClient` instances are running against the same event hub, consumer group and
    checkpointing location, the partitions will be evenly distributed among them.

    To enable load-balancing and persisted checkpoints, checkpoint_store must be set when creating the
    `EventHubConsumerClient`.
    If a checkpoint store is not provided, the checkpoint will be maintained internally in memory.

    An `EventHubConsumerClient` can also receive from a specific partition when you call its method `receive()` or
    `receive_batch()` and specify the partition_id.
    Load-balancing won't work in single-partition mode. But users can still save checkpoints if the checkpoint_store
    is set.

    :param str fully_qualified_namespace: The fully qualified host name for the Event Hubs namespace.
     The namespace format is: `<yournamespace>.servicebus.windows.net`.
    :param str eventhub_name: The path of the specific Event Hub to connect the client to.
    :param str consumer_group: Receive events from the event hub for this consumer group.
    :param credential: The credential object used for authentication which
     implements a particular interface for getting tokens. It accepts
     :class:`EventHubSharedKeyCredential<azure.eventhub.EventHubSharedKeyCredential>`, or credential objects generated
     by the azure-identity library and objects that implement the `get_token(self, *scopes)` method.
    :type credential: ~azure.core.credentials.TokenCredential or ~azure.core.credentials.AzureSasCredential
     or ~azure.core.credentials.AzureNamedKeyCredential
    :keyword bool logging_enable: Whether to output network trace logs to the logger. Default is `False`.
    :keyword float auth_timeout: The time in seconds to wait for a token to be authorized by the service.
     The default value is 60 seconds. If set to 0, no timeout will be enforced from the client.
    :keyword str user_agent: If specified, this will be added in front of the user agent string.
    :keyword retry_total: The total number of attempts to redo a failed operation when an error occurs.
     Default value is 3. The context of `retry_total` in receiving is special: The `receive` method is implemented
     by a while-loop calling internal receive method in each iteration. In the `receive` case,
     `retry_total` specifies the numbers of retry after error raised by internal receive method in the while-loop.
     If retry attempts are exhausted, the `on_error` callback will be called (if provided) with the error information.
     The failed internal partition consumer will be closed (`on_partition_close` will be called if provided) and
     new internal partition consumer will be created (`on_partition_initialize` will be called if provided) to resume
     receiving.
    :paramtype retry_total: int
    :keyword float retry_backoff_factor: A backoff factor to apply between attempts after the second try
     (most errors are resolved immediately by a second try without a delay).
     In fixed mode, retry policy will always sleep for {backoff factor}.
     In 'exponential' mode, retry policy will sleep for: `{backoff factor} * (2 ** ({number of total retries} - 1))`
     seconds. If the backoff_factor is 0.1, then the retry will sleep
     for [0.0s, 0.2s, 0.4s, ...] between retries. The default value is 0.8.
    :keyword float retry_backoff_max: The maximum back off time. Default value is 120 seconds (2 minutes).
    :keyword retry_mode: The delay behavior between retry attempts. Supported values are 'fixed' or 'exponential',
     where default is 'exponential'.
    :paramtype retry_mode: str
    :keyword float idle_timeout: Timeout, in seconds, after which this client will close the underlying connection
     if there is no further activity. By default the value is None, meaning that the client will not shutdown due to
     inactivity unless initiated by the service.
    :keyword transport_type: The type of transport protocol that will be used for communicating with
     the Event Hubs service. Default is `TransportType.Amqp` in which case port 5671 is used.
     If the port 5671 is unavailable/blocked in the network environment, `TransportType.AmqpOverWebsocket` could
     be used instead which uses port 443 for communication.
    :paramtype transport_type: ~azure.eventhub.TransportType
    :keyword dict[str, str or int] http_proxy: HTTP proxy settings. This must be a dictionary with the following
     keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value).
     Additionally the following keys may also be present: `'username', 'password'`.
    :keyword checkpoint_store: A manager that stores the partition load-balancing and checkpoint data
     when receiving events. The checkpoint store will be used in both cases of receiving from all partitions
     or a single partition. In the latter case load-balancing does not apply.
     If a checkpoint store is not provided, the checkpoint will be maintained internally
     in memory, and the `EventHubConsumerClient` instance will receive events without load-balancing.
    :paramtype checkpoint_store: ~azure.eventhub.CheckpointStore or None
    :keyword float load_balancing_interval: When load-balancing kicks in. This is the interval, in seconds,
     between two load-balancing evaluations. Default is 30 seconds.
    :keyword float partition_ownership_expiration_interval: A partition ownership will expire after this number
     of seconds. Every load-balancing evaluation will automatically extend the ownership expiration time.
     Default is 6 * load_balancing_interval, i.e. 180 seconds when using the default load_balancing_interval
     of 30 seconds.
    :keyword load_balancing_strategy: When load-balancing kicks in,
     it will use this strategy to claim and balance the partition ownership.
     Use "greedy" or `LoadBalancingStrategy.GREEDY` for the greedy strategy, which, for every
     load-balancing evaluation, will grab as many unclaimed partitions required to balance the load.
     Use "balanced" or `LoadBalancingStrategy.BALANCED` for the balanced strategy, which, for every load-balancing
     evaluation, claims only one partition that is not claimed by other `EventHubConsumerClient`.
     If all partitions of an EventHub are claimed by other `EventHubConsumerClient` and this client has claimed
     too few partitions, this client will steal one partition from other clients for every load-balancing
     evaluation regardless of the load balancing strategy.
     Greedy strategy is used by default.
    :paramtype load_balancing_strategy: str or ~azure.eventhub.LoadBalancingStrategy
    :keyword custom_endpoint_address: The custom endpoint address to use for establishing a connection to
     the Event Hubs service, allowing network requests to be routed through any application gateways or
     other paths needed for the host environment. Default is None.
     Unless specified otherwise, default transport type is TransportType.AmqpOverWebsockets.
     The format would be like "sb://<custom_endpoint_hostname>:<custom_endpoint_port>".
     If port is not specified in the `custom_endpoint_address`, by default port 443 will be used.
    :paramtype custom_endpoint_address: str or None
    :keyword connection_verify: Path to the custom CA_BUNDLE file of the SSL certificate which is used to
     authenticate the identity of the connection endpoint.
     Default is None in which case `certifi.where()` will be used.
    :paramtype connection_verify: str or None
    :keyword ssl_context: The SSLContext object to use in the underlying Pure Python AMQP transport. If specified,
     connection_verify will be ignored.
    :paramtype ssl_context: ssl.SSLContext or None
    :keyword uamqp_transport: Whether to use the `uamqp` library as the underlying transport. The default value is
     False and the Pure Python AMQP library will be used as the underlying transport.
    :paramtype uamqp_transport: bool
    :keyword float socket_timeout: The time in seconds that the underlying socket on the connection should
     wait when sending and receiving data before timing out. The default value is 0.2 for TransportType.Amqp
     and 1 for TransportType.AmqpOverWebsocket. If EventHubsConnectionError errors are occurring due to write
     timing out, a larger than default value may need to be passed in. This is for advanced usage scenarios
     and ordinarily the default value should be sufficient.

    .. admonition:: Example:

        .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
            :start-after: [START create_eventhub_consumer_client_sync]
            :end-before: [END create_eventhub_consumer_client_sync]
            :language: python
            :dedent: 4
            :caption: Create a new instance of the EventHubConsumerClient.
    """

    def __init__(
        self,
        fully_qualified_namespace: str,
        eventhub_name: str,
        consumer_group: str,
        credential: "CredentialTypes",
        **kwargs: Any,
    ) -> None:
        # Deprecation of uamqp transport
        if kwargs.get("uamqp_transport"):
            warnings.warn(
                "uAMQP legacy support will be removed in the 5.16.0 minor release. "
                "Please remove the use of `uamqp_transport` keyword argument from the client in order "
                "to use the pure Python AMQP transport. "
                "If you rely on this, please comment on [this issue]"
                "(https://github.com/Azure/azure-sdk-for-python/issues/40347) ",
                DeprecationWarning, stacklevel=2
            )
        self._checkpoint_store = kwargs.pop("checkpoint_store", None)
        self._load_balancing_interval = kwargs.pop("load_balancing_interval", None)
        if self._load_balancing_interval is None:
            self._load_balancing_interval = 30
        self._partition_ownership_expiration_interval = kwargs.pop("partition_ownership_expiration_interval", None)
        if self._partition_ownership_expiration_interval is None:
            self._partition_ownership_expiration_interval = 6 * self._load_balancing_interval
        load_balancing_strategy = kwargs.pop("load_balancing_strategy", None)
        self._load_balancing_strategy = (
            LoadBalancingStrategy(load_balancing_strategy) if load_balancing_strategy else LoadBalancingStrategy.GREEDY
        )
        self._consumer_group = consumer_group
        network_tracing = kwargs.pop("logging_enable", False)
        super(EventHubConsumerClient, self).__init__(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential,
            network_tracing=network_tracing,
            **kwargs,
        )
        # consumer auth URI additionally includes consumer group
        self._auth_uri = f"sb://{self._address.hostname}{self._address.path}/consumergroups/{self._consumer_group}"
        self._lock = threading.Lock()
        self._event_processors: Dict[Tuple[str, str], EventProcessor] = {}

    def __enter__(self) -> "EventHubConsumerClient":
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()

    def _create_consumer(
        self,
        consumer_group: str,
        partition_id: str,
        event_position: Union[str, int, datetime.datetime],
        on_event_received: Callable[["PartitionContext", "EventData"], None],
        **kwargs: Any,
    ) -> EventHubConsumer:
        owner_level = kwargs.get("owner_level")
        prefetch = kwargs.get("prefetch") or self._config.prefetch
        track_last_enqueued_event_properties = kwargs.get("track_last_enqueued_event_properties", False)
        event_position_inclusive = kwargs.get("event_position_inclusive", False)

        source_url = "amqps://{}{}/ConsumerGroups/{}/Partitions/{}".format(
            self._address.hostname, self._address.path, consumer_group, partition_id
        )
        handler = EventHubConsumer(
            self,
            source_url,
            event_position=event_position,
            event_position_inclusive=event_position_inclusive,
            owner_level=owner_level,
            on_event_received=on_event_received,
            prefetch=prefetch,
            idle_timeout=self._idle_timeout,
            track_last_enqueued_event_properties=track_last_enqueued_event_properties,
            amqp_transport=self._amqp_transport,
        )
        return handler

    @classmethod
    def from_connection_string(
        cls,
        conn_str: str,
        consumer_group: str,
        *,
        eventhub_name: Optional[str] = None,
        logging_enable: bool = False,
        http_proxy: Optional[Dict[str, Union[str, int]]] = None,
        auth_timeout: float = 60,
        user_agent: Optional[str] = None,
        retry_total: int = 3,
        retry_backoff_factor: float = 0.8,
        retry_backoff_max: float = 120,
        retry_mode: Literal["exponential", "fixed"] = "exponential",
        idle_timeout: Optional[float] = None,
        transport_type: TransportType = TransportType.Amqp,
        checkpoint_store: Optional["CheckpointStore"] = None,
        load_balancing_interval: float = 30,
        partition_ownership_expiration_interval: Optional[float] = None,
        load_balancing_strategy: Union[str, LoadBalancingStrategy] = LoadBalancingStrategy.GREEDY,
        custom_endpoint_address: Optional[str] = None,
        connection_verify: Optional[str] = None,
        ssl_context: Optional["SSLContext"] = None,
        uamqp_transport: bool = False,
        **kwargs: Any,
    ) -> "EventHubConsumerClient":
        """Create an EventHubConsumerClient from a connection string.

        :param str conn_str: The connection string of an Event Hub.
        :param str consumer_group: Receive events from the Event Hub for this consumer group.
        :keyword str eventhub_name: The path of the specific Event Hub to connect the client to.
        :keyword bool logging_enable: Whether to output network trace logs to the logger. Default is `False`.
        :keyword float auth_timeout: The time in seconds to wait for a token to be authorized by the service.
         The default value is 60 seconds. If set to 0, no timeout will be enforced from the client.
        :keyword str user_agent: If specified, this will be added in front of the user agent string.
        :keyword retry_total: The total number of attempts to redo a failed operation when an error occurs.
         Default value is 3. The context of `retry_total` in receiving is special: The `receive` method is implemented
         by a while-loop calling internal receive method in each iteration. In the `receive` case,
         `retry_total` specifies the numbers of retry after error raised by internal receive method in the while-loop.
         If retry attempts are exhausted, the `on_error` callback will be called (if provided) with the error
         information. The failed internal partition consumer will be closed (`on_partition_close` will be called
         if provided) and new internal partition consumer will be created (`on_partition_initialize` will be called if
         provided) to resume receiving.
        :paramtype retry_total: int
        :keyword float retry_backoff_factor: A backoff factor to apply between attempts after the second try
         (most errors are resolved immediately by a second try without a delay).
         In fixed mode, retry policy will always sleep for {backoff factor}.
         In 'exponential' mode, retry policy will sleep for: `{backoff factor} * (2 ** ({number of total retries} - 1))`
         seconds. If the backoff_factor is 0.1, then the retry will sleep
         for [0.0s, 0.2s, 0.4s, ...] between retries. The default value is 0.8.
        :keyword float retry_backoff_max: The maximum back off time. Default value is 120 seconds (2 minutes).
        :keyword retry_mode: The delay behavior between retry attempts. Supported values are 'fixed' or 'exponential',
         where default is 'exponential'.
        :paramtype retry_mode: str
        :keyword float idle_timeout: Timeout, in seconds, after which this client will close the underlying connection
         if there is no furthur activity. By default the value is None, meaning that the client will not shutdown due
         to inactivity unless initiated by the service.
        :keyword transport_type: The type of transport protocol that will be used for communicating with
         the Event Hubs service. Default is `TransportType.Amqp` in which case port 5671 is used.
         If the port 5671 is unavailable/blocked in the network environment, `TransportType.AmqpOverWebsocket` could
         be used instead which uses port 443 for communication.
        :paramtype transport_type: ~azure.eventhub.TransportType
        :keyword dict http_proxy: HTTP proxy settings. This must be a dictionary with the following
         keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value).
         Additionally the following keys may also be present: `'username', 'password'`.
        :keyword checkpoint_store: A manager that stores the partition load-balancing and checkpoint data
         when receiving events. The checkpoint store will be used in both cases of receiving from all partitions
         or a single partition. In the latter case load-balancing does not apply.
         If a checkpoint store is not provided, the checkpoint will be maintained internally
         in memory, and the `EventHubConsumerClient` instance will receive events without load-balancing.
        :paramtype checkpoint_store: ~azure.eventhub.CheckpointStore or None
        :keyword float load_balancing_interval: When load-balancing kicks in. This is the interval, in seconds,
         between two load-balancing evaluations. Default is 10 seconds.
        :keyword float partition_ownership_expiration_interval: A partition ownership will expire after this number
         of seconds. Every load-balancing evaluation will automatically extend the ownership expiration time.
         Default is 6 * load_balancing_interval, i.e. 60 seconds when using the default load_balancing_interval
         of 30 seconds.
        :keyword load_balancing_strategy: When load-balancing kicks in,
         it will use this strategy to claim and balance the partition ownership.
         Use "greedy" or `LoadBalancingStrategy.GREEDY` for the greedy strategy, which, for every
         load-balancing evaluation, will grab as many unclaimed partitions required to balance the load.
         Use "balanced" or `LoadBalancingStrategy.BALANCED` for the balanced strategy, which, for every load-balancing
         evaluation, claims only one partition that is not claimed by other `EventHubConsumerClient`.
         If all partitions of an EventHub are claimed by other `EventHubConsumerClient` and this client has claimed
         too few partitions, this client will steal one partition from other clients for every load-balancing
         evaluation regardless of the load balancing strategy.
         Greedy strategy is used by default.
        :paramtype load_balancing_strategy: str or ~azure.eventhub.LoadBalancingStrategy
        :keyword custom_endpoint_address: The custom endpoint address to use for establishing a connection to
         the Event Hubs service, allowing network requests to be routed through any application gateways or
         other paths needed for the host environment. Default is None.
         Unless specified otherwise, default transport type is TransportType.AmqpOverWebsockets.
         The format would be like "sb://<custom_endpoint_hostname>:<custom_endpoint_port>".
         If port is not specified in the `custom_endpoint_address`, by default port 443 will be used.
        :paramtype custom_endpoint_address: str or None
        :keyword connection_verify: Path to the custom CA_BUNDLE file of the SSL certificate which is used to
         authenticate the identity of the connection endpoint.
         Default is None in which case `certifi.where()` will be used.
        :paramtype connection_verify: str or None
        :keyword ssl_context: The SSLContext object to use in the underlying Pure Python AMQP transport. If specified,
         connection_verify will be ignored.
        :paramtype ssl_context: ssl.SSLContext or None
        :keyword uamqp_transport: Whether to use the `uamqp` library as the underlying transport. The default value is
         False and the Pure Python AMQP library will be used as the underlying transport.
        :paramtype uamqp_transport: bool
        :returns: An EventHubConsumerClient instance.
        :rtype: ~azure.eventhub.EventHubConsumerClient


        .. admonition:: Example:

            .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
                :start-after: [START create_eventhub_consumer_client_from_conn_str_sync]
                :end-before: [END create_eventhub_consumer_client_from_conn_str_sync]
                :language: python
                :dedent: 4
                :caption: Create a new instance of the EventHubConsumerClient from connection string.

        """
        constructor_args = cls._from_connection_string(
            conn_str,
            consumer_group=consumer_group,
            eventhub_name=eventhub_name,
            logging_enable=logging_enable,
            http_proxy=http_proxy,
            auth_timeout=auth_timeout,
            user_agent=user_agent,
            retry_total=retry_total,
            retry_backoff_factor=retry_backoff_factor,
            retry_backoff_max=retry_backoff_max,
            retry_mode=retry_mode,
            idle_timeout=idle_timeout,
            transport_type=transport_type,
            checkpoint_store=checkpoint_store,
            load_balancing_interval=load_balancing_interval,
            partition_ownership_expiration_interval=partition_ownership_expiration_interval,
            load_balancing_strategy=load_balancing_strategy,
            custom_endpoint_address=custom_endpoint_address,
            connection_verify=connection_verify,
            ssl_context=ssl_context,
            uamqp_transport=uamqp_transport,
            **kwargs,
        )
        return cls(**constructor_args)

    @overload
    def _receive(
        self, on_event: Callable[["PartitionContext", Optional["EventData"]], None], **kwargs: Any
    ) -> None: ...

    @overload
    def _receive(self, on_event: Callable[["PartitionContext", List["EventData"]], None], **kwargs: Any) -> None: ...

    def _receive(  # pylint:disable=unused-argument
        self,
        on_event: Union[
            Callable[["PartitionContext", Optional["EventData"]], None],
            Callable[["PartitionContext", List["EventData"]], None],
        ],
        *,
        batch: bool = False,
        max_wait_time: Optional[float] = None,
        partition_id: Optional[str] = None,
        max_batch_size: int = 300,
        owner_level: Optional[int] = None,
        prefetch: int = 300,
        track_last_enqueued_event_properties: bool = False,
        starting_position: Optional[Union[str, int, datetime.datetime, Dict[str, Any]]] = None,
        starting_position_inclusive: Union[bool, Dict[str, bool]] = False,
        on_error: Optional[Callable[["PartitionContext", Exception], None]] = None,
        on_partition_initialize: Optional[Callable[["PartitionContext"], None]] = None,
        on_partition_close: Optional[Callable[["PartitionContext", "CloseReason"], None]] = None,
        **kwargs: Any,
    ) -> None:
        with self._lock:
            error: Optional[str] = None
            if (self._consumer_group, ALL_PARTITIONS) in self._event_processors:
                error = (
                    "This consumer client is already receiving events "
                    "from all partitions for consumer group {}.".format(self._consumer_group)
                )
            elif partition_id is None and any(x[0] == self._consumer_group for x in self._event_processors):
                error = "This consumer client is already receiving events " "for consumer group {}.".format(
                    self._consumer_group
                )
            elif (self._consumer_group, partition_id) in self._event_processors:
                error = (
                    "This consumer client is already receiving events "
                    "from partition {} for consumer group {}. ".format(partition_id, self._consumer_group)
                )
            if error:
                _LOGGER.warning(error)
                raise ValueError(error)

            initial_event_position = starting_position if starting_position is not None else "@latest"
            initial_event_position_inclusive = starting_position_inclusive or False
            event_processor = EventProcessor(
                self,
                self._consumer_group,
                on_event,
                batch=batch,
                max_batch_size=max_batch_size,
                partition_id=partition_id,
                checkpoint_store=self._checkpoint_store,
                load_balancing_interval=self._load_balancing_interval,
                load_balancing_strategy=self._load_balancing_strategy,
                partition_ownership_expiration_interval=self._partition_ownership_expiration_interval,
                initial_event_position=initial_event_position,
                initial_event_position_inclusive=initial_event_position_inclusive,
                max_wait_time=max_wait_time,
                owner_level=owner_level,
                prefetch=prefetch,
                track_last_enqueued_event_properties=track_last_enqueued_event_properties,
                on_error=on_error,
                on_partition_initialize=on_partition_initialize,
                on_partition_close=on_partition_close,
            )
            self._event_processors[(self._consumer_group, partition_id or ALL_PARTITIONS)] = event_processor
        try:
            event_processor.start()
        finally:
            event_processor.stop()
            with self._lock:
                try:
                    del self._event_processors[(self._consumer_group, partition_id or ALL_PARTITIONS)]
                except KeyError:
                    pass

    def receive(  # pylint:disable=unused-argument
        self,
        on_event: Callable[["PartitionContext", Optional["EventData"]], None],
        *,
        max_wait_time: Optional[float] = None,
        partition_id: Optional[str] = None,
        owner_level: Optional[int] = None,
        prefetch: int = 300,
        track_last_enqueued_event_properties: bool = False,
        starting_position: Optional[Union[str, int, datetime.datetime, Dict[str, Any]]] = None,
        starting_position_inclusive: Union[bool, Dict[str, bool]] = False,
        on_error: Optional[Callable[["PartitionContext", Exception], None]] = None,
        on_partition_initialize: Optional[Callable[["PartitionContext"], None]] = None,
        on_partition_close: Optional[Callable[["PartitionContext", "CloseReason"], None]] = None,
        **kwargs: Any,
    ) -> None:
        """Receive events from partition(s), with optional load-balancing and checkpointing.

        :param on_event: The callback function for handling a received event. The callback takes two
         parameters: `partition_context` which contains partition context and `event` which is the received event.
         The callback function should be defined like: `on_event(partition_context, event)`.
         For detailed partition context information, please refer to
         :class:`PartitionContext<azure.eventhub.PartitionContext>`.
        :type on_event: callable[~azure.eventhub.PartitionContext, ~azure.eventhub.EventData or None]
        :keyword float max_wait_time: The maximum interval in seconds that the event processor will wait before calling
         the callback. If no events are received within this interval, the `on_event` callback will be called with
         `None`.
         If this value is set to `None` or 0 (the default), the callback will not be called until an event is received.
        :keyword str partition_id: If specified, the client will receive from this partition only.
         Otherwise the client will receive from all partitions.
        :keyword int owner_level: The priority for an exclusive consumer. An exclusive
         consumer will be created if owner_level is set. A consumer with a higher owner_level has higher exclusive
         priority. The owner level is also know as the 'epoch value' of the consumer.
        :keyword int prefetch: The number of events to prefetch from the service
         for processing. Default is 300.
        :keyword bool track_last_enqueued_event_properties: Indicates whether the consumer should request information
         on the last-enqueued event on its associated partition, and track that information as events are received.
         When information about the partitions last-enqueued event is being tracked, each event received from the
         Event Hubs service will carry metadata about the partition. This results in a small amount of additional
         network bandwidth consumption that is generally a favorable trade-off when considered against periodically
         making requests for partition properties using the Event Hub client.
         It is set to `False` by default.
        :keyword starting_position: Start receiving from this event position
         if there is no checkpoint data for a partition. Checkpoint data will be used if available. This can be a
         a dict with partition ID as the key and position as the value for individual partitions, or a single
         value for all partitions. The value type can be str, int or datetime.datetime. Also supported are the
         values "-1" for receiving from the beginning of the stream, and "@latest" for receiving only new events.
         Default value is "@latest".
        :paramtype starting_position: str, int, datetime.datetime or dict[str,any]
        :keyword starting_position_inclusive: Determine whether the given starting_position is inclusive(>=) or
         not (>). True for inclusive and False for exclusive. This can be a dict with partition ID as the key and
         bool as the value indicating whether the starting_position for a specific partition is inclusive or not.
         This can also be a single bool value for all starting_position. The default value is False.
        :paramtype starting_position_inclusive: bool or dict[str,bool]
        :keyword on_error: The callback function that will be called when an error is raised during receiving
         after retry attempts are exhausted, or during the process of load-balancing.
         The callback takes two parameters: `partition_context` which contains partition information
         and `error` being the exception. `partition_context` could be None if the error is raised during
         the process of load-balance. The callback should be defined like: `on_error(partition_context, error)`.
         The `on_error` callback will also be called if an unhandled exception is raised during
         the `on_event` callback.
        :paramtype on_error: callable[[~azure.eventhub.PartitionContext, Exception]]
        :keyword on_partition_initialize: The callback function that will be called after a consumer for a certain
         partition finishes initialization. It would also be called when a new internal partition consumer is created
         to take over the receiving process for a failed and closed internal partition consumer.
         The callback takes a single parameter: `partition_context`
         which contains the partition information. The callback should be defined
         like: `on_partition_initialize(partition_context)`.
        :paramtype on_partition_initialize: callable[[~azure.eventhub.PartitionContext]]
        :keyword on_partition_close: The callback function that will be called after a consumer for a certain
         partition is closed. It would be also called when error is raised during receiving after retry attempts are
         exhausted. The callback takes two parameters: `partition_context` which contains partition
         information and `reason` for the close. The callback should be defined like:
         `on_partition_close(partition_context, reason)`.
         Please refer to :class:`CloseReason<azure.eventhub.CloseReason>` for the various closing reasons.
        :paramtype on_partition_close: callable[[~azure.eventhub.PartitionContext, ~azure.eventhub.CloseReason]]
        :rtype: None

        .. admonition:: Example:

            .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
                :start-after: [START eventhub_consumer_client_receive_sync]
                :end-before: [END eventhub_consumer_client_receive_sync]
                :language: python
                :dedent: 4
                :caption: Receive events from the EventHub.
        """
        self._receive(
            on_event,
            batch=False,
            max_batch_size=1,
            max_wait_time=max_wait_time,
            partition_id=partition_id,
            owner_level=owner_level,
            prefetch=prefetch,
            track_last_enqueued_event_properties=track_last_enqueued_event_properties,
            starting_position=starting_position,
            starting_position_inclusive=starting_position_inclusive,
            on_error=on_error,
            on_partition_initialize=on_partition_initialize,
            on_partition_close=on_partition_close,
        )

    def receive_batch(
        self,
        on_event_batch: Callable[["PartitionContext", List["EventData"]], None],
        *,
        max_batch_size: int = 300,
        max_wait_time: Optional[float] = None,
        partition_id: Optional[str] = None,
        owner_level: Optional[int] = None,
        prefetch: int = 300,
        track_last_enqueued_event_properties: bool = False,
        starting_position: Optional[Union[str, int, datetime.datetime, Dict[str, Any]]] = None,
        starting_position_inclusive: Union[bool, Dict[str, bool]] = False,
        on_error: Optional[Callable[["PartitionContext", Exception], None]] = None,
        on_partition_initialize: Optional[Callable[["PartitionContext"], None]] = None,
        on_partition_close: Optional[Callable[["PartitionContext", "CloseReason"], None]] = None,
        **kwargs: Any,
    ) -> None:
        """Receive events from partition(s), with optional load-balancing and checkpointing.

        :param on_event_batch: The callback function for handling a batch of received events. The callback takes two
         parameters: `partition_context` which contains partition context and `event_batch`, which is the received
         events. The callback function should be defined like: `on_event_batch(partition_context, event_batch)`.
         `event_batch` could be an empty list if `max_wait_time` is not None nor 0 and no event is received
         after `max_wait_time`.
         For detailed partition context information, please refer to
         :class:`PartitionContext<azure.eventhub.PartitionContext>`.
        :type on_event_batch: callable[~azure.eventhub.PartitionContext, list[~azure.eventhub.EventData]]
        :keyword int max_batch_size: The maximum number of events in a batch passed to callback `on_event_batch`.
         If the actual received number of events is larger than `max_batch_size`, the received events are divided into
         batches and call the callback for each batch with up to `max_batch_size` events.
        :keyword float max_wait_time: The maximum interval in seconds that the event processor will wait before
         calling the callback.
         If no events are received within this interval, the `on_event_batch` callback will be called
         with an empty list.
        :keyword str partition_id: If specified, the client will receive from this partition only.
         Otherwise the client will receive from all partitions.
        :keyword int owner_level: The priority for an exclusive consumer. An exclusive
         consumer will be created if owner_level is set. A consumer with a higher owner_level has higher exclusive
         priority. The owner level is also know as the 'epoch value' of the consumer.
        :keyword int prefetch: The number of events to prefetch from the service
         for processing. Default is 300.
        :keyword bool track_last_enqueued_event_properties: Indicates whether the consumer should request information
         on the last-enqueued event on its associated partition, and track that information as events are received.
         When information about the partitions last-enqueued event is being tracked, each event received from the
         Event Hubs service will carry metadata about the partition. This results in a small amount of additional
         network bandwidth consumption that is generally a favorable trade-off when considered against periodically
         making requests for partition properties using the Event Hub client.
         It is set to `False` by default.
        :keyword starting_position: Start receiving from this event position
         if there is no checkpoint data for a partition. Checkpoint data will be used if available. This can be a
         a dict with partition ID as the key and position as the value for individual partitions, or a single
         value for all partitions. The value type can be str, int or datetime.datetime. Also supported are the
         values "-1" for receiving from the beginning of the stream, and "@latest" for receiving only new events.
         Default value is "@latest".
        :paramtype starting_position: str, int, datetime.datetime or dict[str,any]
        :keyword starting_position_inclusive: Determine whether the given starting_position is inclusive(>=) or
         not (>). True for inclusive and False for exclusive. This can be a dict with partition ID as the key and
         bool as the value indicating whether the starting_position for a specific partition is inclusive or not.
         This can also be a single bool value for all starting_position. The default value is False.
        :paramtype starting_position_inclusive: bool or dict[str,bool]
        :keyword on_error: The callback function that will be called when an error is raised during receiving
         after retry attempts are exhausted, or during the process of load-balancing.
         The callback takes two parameters: `partition_context` which contains partition information
         and `error` being the exception. `partition_context` could be None if the error is raised during
         the process of load-balance. The callback should be defined like: `on_error(partition_context, error)`.
         The `on_error` callback will also be called if an unhandled exception is raised during
         the `on_event` callback.
        :paramtype on_error: callable[[~azure.eventhub.PartitionContext, Exception]]
        :keyword on_partition_initialize: The callback function that will be called after a consumer for a certain
         partition finishes initialization. It would also be called when a new internal partition consumer is created
         to take over the receiving process for a failed and closed internal partition consumer.
         The callback takes a single parameter: `partition_context`
         which contains the partition information. The callback should be defined
         like: `on_partition_initialize(partition_context)`.
        :paramtype on_partition_initialize: callable[[~azure.eventhub.PartitionContext]]
        :keyword on_partition_close: The callback function that will be called after a consumer for a certain
         partition is closed. It would be also called when error is raised during receiving after retry attempts are
         exhausted. The callback takes two parameters: `partition_context` which contains partition
         information and `reason` for the close. The callback should be defined like:
         `on_partition_close(partition_context, reason)`.
         Please refer to :class:`CloseReason<azure.eventhub.CloseReason>` for the various closing reasons.
        :paramtype on_partition_close: callable[[~azure.eventhub.PartitionContext, ~azure.eventhub.CloseReason]]
        :rtype: None

        .. admonition:: Example:

            .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
                :start-after: [START eventhub_consumer_client_receive_batch_sync]
                :end-before: [END eventhub_consumer_client_receive_batch_sync]
                :language: python
                :dedent: 4
                :caption: Receive events in batches from the EventHub.
        """
        self._receive(
            on_event_batch,
            batch=True,
            max_batch_size=max_batch_size,
            max_wait_time=max_wait_time,
            partition_id=partition_id,
            owner_level=owner_level,
            prefetch=prefetch,
            track_last_enqueued_event_properties=track_last_enqueued_event_properties,
            starting_position=starting_position,
            starting_position_inclusive=starting_position_inclusive,
            on_error=on_error,
            on_partition_initialize=on_partition_initialize,
            on_partition_close=on_partition_close,
            **kwargs,
        )

    def get_eventhub_properties(self) -> Dict[str, Any]:
        """Get properties of the Event Hub.

        Keys in the returned dictionary include:

            - `eventhub_name` (str)
            - `created_at` (UTC datetime.datetime)
            - `partition_ids` (list[str])

        :return: A dictionary containing information about the Event Hub.
        :rtype: dict[str, any]
        :raises: :class:`EventHubError<azure.eventhub.exceptions.EventHubError>`
        """
        return super(EventHubConsumerClient, self)._get_eventhub_properties()

    def get_partition_ids(self) -> List[str]:
        """Get partition IDs of the Event Hub.

        :return: A list of partition IDs.
        :rtype: list[str]
        :raises: :class:`EventHubError<azure.eventhub.exceptions.EventHubError>`
        """
        return super(EventHubConsumerClient, self)._get_partition_ids()

    def get_partition_properties(self, partition_id: str) -> Dict[str, Any]:
        """Get properties of the specified partition.

        Keys in the properties dictionary include:

            - `eventhub_name` (str)
            - `id` (str)
            - `beginning_sequence_number` (int)
            - `last_enqueued_sequence_number` (int)
            - `last_enqueued_offset` (str)
            - `last_enqueued_time_utc` (UTC datetime.datetime)
            - `is_empty` (bool)

        :param partition_id: The target partition ID.
        :type partition_id: str
        :return: A dictionary containing partition properties.
        :rtype: dict[str, any]
        :raises: :class:`EventHubError<azure.eventhub.exceptions.EventHubError>`
        """
        return super(EventHubConsumerClient, self)._get_partition_properties(partition_id)

    def close(self) -> None:
        """Stop retrieving events from the Event Hub and close the underlying AMQP connection and links.

        :rtype: None

        .. admonition:: Example:

            .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
                :start-after: [START eventhub_consumer_client_close_sync]
                :end-before: [END eventhub_consumer_client_close_sync]
                :language: python
                :dedent: 4
                :caption: Close down the client.

        """
        with self._lock:
            for processor in self._event_processors.values():
                processor.stop()
            self._event_processors = {}
        super(EventHubConsumerClient, self)._close()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# -----------------------------------------------------------------------------------
from typing import Dict, Any, Iterable, Optional, Union
import time
import uuid
import logging

from .checkpoint_store import CheckpointStore

_LOGGER = logging.getLogger(__name__)


class _TrieNode:
    def __init__(self, value, is_leaf):
        self.value = value
        self.is_leaf = is_leaf
        if is_leaf:
            self.children = None
        else:
            self.children = {}

    def __contains__(self, item):
        if self.is_leaf:
            raise ValueError("Please don't call __contains__ on a leaf TrieNode")
        return item in self.children

    def get(self, key):
        if self.is_leaf:
            raise ValueError("Please don't get a child from a leaf TrieNode")
        return self.children.get(key)

    def put(self, key, value):
        if self.is_leaf:
            raise ValueError("Please don't put a value to a leaf TrieNode")
        self.children[key] = value


def _lookup_trie(root, path, path_index):
    if path_index == len(path) - 1:
        return root.get(path[path_index])
    if path[path_index] in root:
        return _lookup_trie(root.children[path[path_index]], path, path_index + 1)
    return None


def _list_leaves_trie(root, result):
    if root.is_leaf:
        result.append(root.value)
        return
    for child in root.children.values():
        _list_leaves_trie(child, result)


def _set_ele_trie(root, ele, keys_path, path_index):
    if path_index == len(keys_path) - 1:
        root.children[ele[keys_path[path_index]]] = _TrieNode(ele, True)
        return
    next_node = root.get(ele[keys_path[path_index]])
    if not next_node:
        next_node = _TrieNode(ele[keys_path[path_index]], False)
        root.put(ele[keys_path[path_index]], next_node)
    _set_ele_trie(next_node, ele, keys_path, path_index + 1)


class _DictTrie:
    def __init__(self, root_name, keys_path):
        self._root = _TrieNode(root_name, False)
        self._keys_path = keys_path

    def lookup(self, path):
        return _lookup_trie(self._root, path, 0)

    def list_leaves(self, node):
        if not node:
            node = self._root
        result = []
        _list_leaves_trie(node, result)
        return result

    def set_ele(self, ele):
        _set_ele_trie(self._root, ele, self._keys_path, 0)


class InMemoryCheckpointStore(CheckpointStore):
    def __init__(self):
        self._ownerships_trie = _DictTrie(
            "ownerships_trie",
            keys_path=(
                "fully_qualified_namespace",
                "eventhub_name",
                "consumer_group",
                "partition_id",
            ),
        )
        self._checkpoints_trie = _DictTrie(
            "checkpoints_trie",
            keys_path=(
                "fully_qualified_namespace",
                "eventhub_name",
                "consumer_group",
                "partition_id",
            ),
        )

    def list_ownership(
        self, fully_qualified_namespace: str, eventhub_name: str, consumer_group: str, **kwargs: Any
    ) -> Iterable[Dict[str, Any]]:
        consumer_group_node = self._ownerships_trie.lookup((fully_qualified_namespace, eventhub_name, consumer_group))
        return self._ownerships_trie.list_leaves(consumer_group_node)

    def claim_ownership(self, ownership_list: Iterable[Dict[str, Any]], **kwargs: Any) -> Iterable[Dict[str, Any]]:
        result = []
        for ownership in ownership_list:
            fully_qualified_namespace = ownership["fully_qualified_namespace"]
            eventhub_name = ownership["eventhub_name"]
            consumer_group = ownership["consumer_group"]
            partition_id = ownership["partition_id"]
            etag = ownership.get("etag")
            old_ownership_node = self._ownerships_trie.lookup(
                (fully_qualified_namespace, eventhub_name, consumer_group, partition_id)
            )
            if old_ownership_node:
                old_ownership = old_ownership_node.value
                if etag == old_ownership["etag"]:
                    ownership["etag"] = str(uuid.uuid4())
                    ownership["last_modified_time"] = time.time()
                    old_ownership["etag"] = ownership["etag"]
                    old_ownership["last_modified_time"] = ownership["last_modified_time"]
                    old_ownership["owner_id"] = ownership["owner_id"]
                    result.append(old_ownership)
            else:
                ownership["etag"] = str(uuid.uuid4())
                ownership["last_modified_time"] = time.time()
                self._ownerships_trie.set_ele(ownership)
                result.append(ownership)
        return result

    def update_checkpoint(self, checkpoint: Dict[str, Optional[Union[str, int]]], **kwargs: Any) -> None:
        return self._checkpoints_trie.set_ele(checkpoint)

    def list_checkpoints(self, fully_qualified_namespace, eventhub_name, consumer_group, **kwargs):
        consumer_group_node = self._checkpoints_trie.lookup((fully_qualified_namespace, eventhub_name, consumer_group))
        return self._checkpoints_trie.list_leaves(consumer_group_node)



from azure.eventhub.aio import EventHubProducerClient, EventHubConsumerClient
from azure.eventhub import TransportType


def test_custom_endpoint_async():
    producer = EventHubProducerClient(
        "fake.host.com",
        "fake_eh",
        None,
    )
    assert not producer._config.custom_endpoint_hostname
    assert producer._config.transport_type == TransportType.Amqp
    assert producer._config.connection_port == 5671

    producer = EventHubProducerClient("fake.host.com", "fake_eh", None, custom_endpoint_address="https://12.34.56.78")
    assert producer._config.custom_endpoint_hostname == "12.34.56.78"
    assert producer._config.transport_type == TransportType.AmqpOverWebsocket
    assert producer._config.connection_port == 443

    producer = EventHubProducerClient(
        "fake.host.com", "fake_eh", None, custom_endpoint_address="sb://fake.endpoint.com:443"
    )
    assert producer._config.custom_endpoint_hostname == "fake.endpoint.com"
    assert producer._config.transport_type == TransportType.AmqpOverWebsocket
    assert producer._config.connection_port == 443

    producer = EventHubProducerClient(
        "fake.host.com", "fake_eh", None, custom_endpoint_address="https://fake.endpoint.com:200"
    )
    assert producer._config.custom_endpoint_hostname == "fake.endpoint.com"
    assert producer._config.transport_type == TransportType.AmqpOverWebsocket
    assert producer._config.connection_port == 200

    producer = EventHubProducerClient("fake.host.com", "fake_eh", None, custom_endpoint_address="fake.endpoint.com:200")
    assert producer._config.custom_endpoint_hostname == "fake.endpoint.com"
    assert producer._config.transport_type == TransportType.AmqpOverWebsocket
    assert producer._config.connection_port == 200

    consumer = EventHubConsumerClient(
        "fake.host.com",
        "fake_eh",
        "fake_group",
        None,
    )
    assert not consumer._config.custom_endpoint_hostname
    assert consumer._config.transport_type == TransportType.Amqp
    assert consumer._config.connection_port == 5671

    consumer = EventHubConsumerClient(
        "fake.host.com", "fake_eh", "fake_group", None, custom_endpoint_address="https://12.34.56.78/"
    )
    assert consumer._config.custom_endpoint_hostname == "12.34.56.78"
    assert consumer._config.transport_type == TransportType.AmqpOverWebsocket
    assert consumer._config.connection_port == 443

    consumer = EventHubConsumerClient(
        "fake.host.com", "fake_eh", "fake_group", None, custom_endpoint_address="sb://fake.endpoint.com:443"
    )
    assert consumer._config.custom_endpoint_hostname == "fake.endpoint.com"
    assert consumer._config.transport_type == TransportType.AmqpOverWebsocket
    assert consumer._config.connection_port == 443

    consumer = EventHubConsumerClient(
        "fake.host.com", "fake_eh", "fake_group", None, custom_endpoint_address="https://fake.endpoint.com:200"
    )
    assert consumer._config.custom_endpoint_hostname == "fake.endpoint.com"
    assert consumer._config.transport_type == TransportType.AmqpOverWebsocket
    assert consumer._config.connection_port == 200

    consumer = EventHubConsumerClient(
        "fake.host.com", "fake_eh", "fake_group", None, custom_endpoint_address="fake.endpoint.com:200"
    )
    assert consumer._config.custom_endpoint_hostname == "fake.endpoint.com"
    assert consumer._config.transport_type == TransportType.AmqpOverWebsocket
    assert consumer._config.connection_port == 200


def test_custom_certificate_async():
    producer = EventHubProducerClient("fake.host.com", "fake_eh", None, connection_verify="/usr/bin/local/cert")
    assert producer._config.connection_verify == "/usr/bin/local/cert"

    consumer = EventHubConsumerClient(
        "fake.host.com", "fake_eh", "fake_group", None, connection_verify="D:/local/certfile"
    )
    assert consumer._config.connection_verify == "D:/local/certfile"



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# The messaging layer defines two concrete types (source and target) to be used as the source and target of a
# link. These types are supplied in the source and target fields of the attach frame when establishing or
# resuming link. The source is comprised of an address (which the container of the outgoing Link Endpoint will
# resolve to a Node within that container) coupled with properties which determine:
#
#   - which messages from the sending Node will be sent on the Link
#   - how sending the message affects the state of that message at the sending Node
#   - the behavior of Messages which have been transferred on the Link, but have not yet reached a
#     terminal state at the receiver, when the source is destroyed.

# TODO: fix mypy errors for _code/_definition/__defaults__ (issue #26500)
from collections import namedtuple

from .types import AMQPTypes, FieldDefinition, ObjDefinition
from .constants import FIELD
from .performatives import _CAN_ADD_DOCSTRING


class TerminusDurability(object):
    """Durability policy for a terminus.

    <type name="terminus-durability" class="restricted" source="uint">
        <choice name="none" value="0"/>
        <choice name="configuration" value="1"/>
        <choice name="unsettled-state" value="2"/>
    </type>

    Determines which state of the terminus is held durably.
    """

    #: No Terminus state is retained durably
    NoDurability = 0
    #: Only the existence and configuration of the Terminus is retained durably.
    Configuration = 1
    #: In addition to the existence and configuration of the Terminus, the unsettled state for durable
    #: messages is retained durably.
    UnsettledState = 2


class ExpiryPolicy(object):
    """Expiry policy for a terminus.

    <type name="terminus-expiry-policy" class="restricted" source="symbol">
        <choice name="link-detach" value="link-detach"/>
        <choice name="session-end" value="session-end"/>
        <choice name="connection-close" value="connection-close"/>
        <choice name="never" value="never"/>
    </type>

    Determines when the expiry timer of a terminus starts counting down from the timeout
    value. If the link is subsequently re-attached before the terminus is expired, then the
    count down is aborted. If the conditions for the terminus-expiry-policy are subsequently
    re-met, the expiry timer restarts from its originally configured timeout value.
    """

    #: The expiry timer starts when Terminus is detached.
    LinkDetach = b"link-detach"
    #: The expiry timer starts when the most recently associated session is ended.
    SessionEnd = b"session-end"
    #: The expiry timer starts when most recently associated connection is closed.
    ConnectionClose = b"connection-close"
    #: The Terminus never expires.
    Never = b"never"


class DistributionMode(object):
    """Link distribution policy.

    <type name="std-dist-mode" class="restricted" source="symbol" provides="distribution-mode">
        <choice name="move" value="move"/>
        <choice name="copy" value="copy"/>
    </type>

    Policies for distributing messages when multiple links are connected to the same node.
    """

    #: Once successfully transferred over the link, the message will no longer be available
    #: to other links from the same node.
    Move = b"move"
    #: Once successfully transferred over the link, the message is still available for other
    #: links from the same node.
    Copy = b"copy"


class LifeTimePolicy(object):
    #: Lifetime of dynamic node scoped to lifetime of link which caused creation.
    #: A node dynamically created with this lifetime policy will be deleted at the point that the link
    #: which caused its creation ceases to exist.
    DeleteOnClose = 0x0000002B
    #: Lifetime of dynamic node scoped to existence of links to the node.
    #: A node dynamically created with this lifetime policy will be deleted at the point that there remain
    #: no links for which the node is either the source or target.
    DeleteOnNoLinks = 0x0000002C
    #: Lifetime of dynamic node scoped to existence of messages on the node.
    #: A node dynamically created with this lifetime policy will be deleted at the point that the link which
    #: caused its creation no longer exists and there remain no messages at the node.
    DeleteOnNoMessages = 0x0000002D
    #: Lifetime of node scoped to existence of messages on or links to the node.
    #: A node dynamically created with this lifetime policy will be deleted at the point that the there are no
    #: links which have this node as their source or target, and there remain no messages at the node.
    DeleteOnNoLinksOrMessages = 0x0000002E


class SupportedOutcomes(object):
    #: Indicates successful processing at the receiver.
    accepted = b"amqp:accepted:list"
    #: Indicates an invalid and unprocessable message.
    rejected = b"amqp:rejected:list"
    #: Indicates that the message was not (and will not be) processed.
    released = b"amqp:released:list"
    #: Indicates that the message was modified, but not processed.
    modified = b"amqp:modified:list"


class ApacheFilters(object):
    #: Exact match on subject - analogous to legacy AMQP direct exchange bindings.
    legacy_amqp_direct_binding = b"apache.org:legacy-amqp-direct-binding:string"
    #: Pattern match on subject - analogous to legacy AMQP topic exchange bindings.
    legacy_amqp_topic_binding = b"apache.org:legacy-amqp-topic-binding:string"
    #: Matching on message headers - analogous to legacy AMQP headers exchange bindings.
    legacy_amqp_headers_binding = b"apache.org:legacy-amqp-headers-binding:map"
    #: Filter out messages sent from the same connection as the link is currently associated with.
    no_local_filter = b"apache.org:no-local-filter:list"
    #: SQL-based filtering syntax.
    selector_filter = b"apache.org:selector-filter:string"


Source = namedtuple(
    "Source",
    [
        "address",
        "durable",
        "expiry_policy",
        "timeout",
        "dynamic",
        "dynamic_node_properties",
        "distribution_mode",
        "filters",
        "default_outcome",
        "outcomes",
        "capabilities",
    ],
    defaults=(None,) * 11,  # type: ignore
)
Source._code = 0x00000028  # type: ignore # pylint: disable=protected-access
Source._definition = (  # type: ignore # pylint: disable=protected-access
    FIELD("address", AMQPTypes.string, False, None, False),
    FIELD("durable", AMQPTypes.uint, False, "none", False),
    FIELD("expiry_policy", AMQPTypes.symbol, False, ExpiryPolicy.SessionEnd, False),
    FIELD("timeout", AMQPTypes.uint, False, 0, False),
    FIELD("dynamic", AMQPTypes.boolean, False, False, False),
    FIELD("dynamic_node_properties", FieldDefinition.node_properties, False, None, False),
    FIELD("distribution_mode", AMQPTypes.symbol, False, None, False),
    FIELD("filters", FieldDefinition.filter_set, False, None, False),
    FIELD("default_outcome", ObjDefinition.delivery_state, False, None, False),
    FIELD("outcomes", AMQPTypes.symbol, False, None, True),
    FIELD("capabilities", AMQPTypes.symbol, False, None, True),
)
if _CAN_ADD_DOCSTRING:
    Source.__doc__ = """
    For containers which do not implement address resolution (and do not admit spontaneous link
    attachment from their partners) but are instead only used as producers of messages, it is unnecessary to provide
    spurious detail on the source. For this purpose it is possible to use a "minimal" source in which all the
    fields are left unset.

    :param str address: The address of the source.
        The address of the source MUST NOT be set when sent on a attach frame sent by the receiving Link Endpoint
        where the dynamic fiag is set to true (that is where the receiver is requesting the sender to create an
        addressable node). The address of the source MUST be set when sent on a attach frame sent by the sending
        Link Endpoint where the dynamic fiag is set to true (that is where the sender has created an addressable
        node at the request of the receiver and is now communicating the address of that created node).
        The generated name of the address SHOULD include the link name and the container-id of the remote container
        to allow for ease of identification.
    :param ~uamqp.endpoints.TerminusDurability durable: Indicates the durability of the terminus.
        Indicates what state of the terminus will be retained durably: the state of durable messages, only
        existence and configuration of the terminus, or no state at all.
    :param ~uamqp.endpoints.ExpiryPolicy expiry_policy: The expiry policy of the Source.
        Determines when the expiry timer of a Terminus starts counting down from the timeout value. If the link
        is subsequently re-attached before the Terminus is expired, then the count down is aborted. If the
        conditions for the terminus-expiry-policy are subsequently re-met, the expiry timer restarts from its
        originally configured timeout value.
    :param int timeout: Duration that an expiring Source will be retained in seconds.
        The Source starts expiring as indicated by the expiry-policy.
    :param bool dynamic: Request dynamic creation of a remote Node.
        When set to true by the receiving Link endpoint, this field constitutes a request for the sending peer
        to dynamically create a Node at the source. In this case the address field MUST NOT be set. When set to
        true by the sending Link Endpoint this field indicates creation of a dynamically created Node. In this case
        the address field will contain the address of the created Node. The generated address SHOULD include the
        Link name and Session-name or client-id in some recognizable form for ease of traceability.
    :param dict dynamic_node_properties: Properties of the dynamically created Node.
        If the dynamic field is not set to true this field must be left unset. When set by the receiving Link
        endpoint, this field contains the desired properties of the Node the receiver wishes to be created. When
        set by the sending Link endpoint this field contains the actual properties of the dynamically created node.
    :param uamqp.endpoints.DistributionMode distribution_mode: The distribution mode of the Link.
        This field MUST be set by the sending end of the Link if the endpoint supports more than one
        distribution-mode. This field MAY be set by the receiving end of the Link to indicate a preference when a
        Node supports multiple distribution modes.
    :param dict filters: A set of predicates to filter the Messages admitted onto the Link.
        The receiving endpoint sets its desired filter, the sending endpoint sets the filter actually in place
        (including any filters defaulted at the node). The receiving endpoint MUST check that the filter in place
        meets its needs and take responsibility for detaching if it does not.
         Common filter types, along with the capabilities they are associated with are registered
         here: http://www.amqp.org/specification/1.0/filters.
    :param ~uamqp.outcomes.DeliveryState default_outcome: Default outcome for unsettled transfers.
        Indicates the outcome to be used for transfers that have not reached a terminal state at the receiver
        when the transfer is settled, including when the Source is destroyed. The value MUST be a valid
        outcome (e.g. Released or Rejected).
    :param list(bytes) outcomes: Descriptors for the outcomes that can be chosen on this link.
        The values in this field are the symbolic descriptors of the outcomes that can be chosen on this link.
        This field MAY be empty, indicating that the default-outcome will be assumed for all message transfers
        (if the default-outcome is not set, and no outcomes are provided, then the accepted outcome must be
        supported by the source). When present, the values MUST be a symbolic descriptor of a valid outcome,
        e.g. "amqp:accepted:list".
    :param list(bytes) capabilities: The extension capabilities the sender supports/desires.
        See http://www.amqp.org/specification/1.0/source-capabilities.
    """


Target = namedtuple(
    "Target",
    ["address", "durable", "expiry_policy", "timeout", "dynamic", "dynamic_node_properties", "capabilities"],
    defaults=(None,) * 7,  # type: ignore
)
Target._code = 0x00000029  # type: ignore # pylint: disable=protected-access
Target._definition = (  # type: ignore # pylint: disable=protected-access
    FIELD("address", AMQPTypes.string, False, None, False),
    FIELD("durable", AMQPTypes.uint, False, "none", False),
    FIELD("expiry_policy", AMQPTypes.symbol, False, ExpiryPolicy.SessionEnd, False),
    FIELD("timeout", AMQPTypes.uint, False, 0, False),
    FIELD("dynamic", AMQPTypes.boolean, False, False, False),
    FIELD("dynamic_node_properties", FieldDefinition.node_properties, False, None, False),
    FIELD("capabilities", AMQPTypes.symbol, False, None, True),
)
if _CAN_ADD_DOCSTRING:
    Target.__doc__ = """
    For containers which do not implement address resolution (and do not admit spontaneous link attachment
    from their partners) but are instead only used as consumers of messages, it is unnecessary to provide spurious
    detail on the source. For this purpose it is possible to use a 'minimal' target in which all the
    fields are left unset.

    :param str address: The address of the source.
        The address of the source MUST NOT be set when sent on a attach frame sent by the receiving Link Endpoint
        where the dynamic fiag is set to true (that is where the receiver is requesting the sender to create an
        addressable node). The address of the source MUST be set when sent on a attach frame sent by the sending
        Link Endpoint where the dynamic fiag is set to true (that is where the sender has created an addressable
        node at the request of the receiver and is now communicating the address of that created node).
        The generated name of the address SHOULD include the link name and the container-id of the remote container
        to allow for ease of identification.
    :param ~uamqp.endpoints.TerminusDurability durable: Indicates the durability of the terminus.
        Indicates what state of the terminus will be retained durably: the state of durable messages, only
        existence and configuration of the terminus, or no state at all.
    :param ~uamqp.endpoints.ExpiryPolicy expiry_policy: The expiry policy of the Source.
        Determines when the expiry timer of a Terminus starts counting down from the timeout value. If the link
        is subsequently re-attached before the Terminus is expired, then the count down is aborted. If the
        conditions for the terminus-expiry-policy are subsequently re-met, the expiry timer restarts from its
        originally configured timeout value.
    :param int timeout: Duration that an expiring Source will be retained in seconds.
        The Source starts expiring as indicated by the expiry-policy.
    :param bool dynamic: Request dynamic creation of a remote Node.
        When set to true by the receiving Link endpoint, this field constitutes a request for the sending peer
        to dynamically create a Node at the source. In this case the address field MUST NOT be set. When set to
        true by the sending Link Endpoint this field indicates creation of a dynamically created Node. In this case
        the address field will contain the address of the created Node. The generated address SHOULD include the
        Link name and Session-name or client-id in some recognizable form for ease of traceability.
    :param dict dynamic_node_properties: Properties of the dynamically created Node.
        If the dynamic field is not set to true this field must be left unset. When set by the receiving Link
        endpoint, this field contains the desired properties of the Node the receiver wishes to be created. When
        set by the sending Link endpoint this field contains the actual properties of the dynamically created node.
    :param list(bytes) capabilities: The extension capabilities the sender supports/desires.
        See http://www.amqp.org/specification/1.0/source-capabilities.
    """



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
This sample demonstrates how to convert an Iot Hub connection string to
an Event Hubs connection string that points to the built-in messaging endpoint.
The Event Hubs connection string is then used with the EventHubConsumerClient to
receive events.
More information about the built-in messaging endpoint can be found at:
https://learn.microsoft.com/azure/iot-hub/iot-hub-devguide-messages-read-builtin
"""
from typing import cast
import os
import re
import time
from base64 import b64encode, b64decode
from hashlib import sha256
from hmac import HMAC
import asyncio
from urllib.parse import urlencode, quote_plus
from uamqp import ReceiveClient, Source
from uamqp.errors import LinkRedirect

from azure.eventhub.aio import EventHubConsumerClient


def generate_sas_token(uri, policy, key, expiry=None):
    """Create a shared access signiture token as a string literal.
    :returns: SAS token as string literal.
    :rtype: str
    """
    if not expiry:
        expiry = time.time() + 3600
    encoded_uri = quote_plus(uri)
    ttl = int(expiry)
    sign_key = "%s\n%d" % (encoded_uri, ttl)
    signature = b64encode(HMAC(b64decode(key), sign_key.encode("utf-8"), sha256).digest())
    result = {"sr": uri, "sig": signature, "se": str(ttl)}
    if policy:
        result["skn"] = policy
    return "SharedAccessSignature " + urlencode(result)


def parse_iot_conn_str(iothub_conn_str):
    hostname = None
    shared_access_key_name = None
    shared_access_key = None
    for element in iothub_conn_str.split(";"):
        key, _, value = element.partition("=")
        if key.lower() == "hostname":
            hostname = value.rstrip("/")
        elif key.lower() == "sharedaccesskeyname":
            shared_access_key_name = value
        elif key.lower() == "sharedaccesskey":
            shared_access_key = value
    if not all([hostname, shared_access_key_name, shared_access_key]):
        raise ValueError("Invalid connection string")
    return hostname, shared_access_key_name, shared_access_key


def convert_iothub_to_eventhub_conn_str(iothub_conn_str):
    hostname, shared_access_key_name, shared_access_key = parse_iot_conn_str(iothub_conn_str)
    iot_hub_name = hostname.split(".")[0]
    operation = "/messages/events/ConsumerGroups/{}/Partitions/{}".format("$Default", 0)
    username = "{}@sas.root.{}".format(shared_access_key_name, iot_hub_name)
    sas_token = generate_sas_token(hostname, shared_access_key_name, shared_access_key)
    uri = "amqps://{}:{}@{}{}".format(quote_plus(username), quote_plus(sas_token), hostname, operation)
    source_uri = Source(uri)
    receive_client = ReceiveClient(source_uri)
    try:
        receive_client.receive_message_batch(max_batch_size=1)
    except LinkRedirect as redirect:
        # Once a redirect error is received, close the original client and recreate a new one to the re-directed address
        receive_client.close()
        fully_qualified_name = redirect.hostname.decode("utf-8")
        # Use regular expression to parse the Event Hub name from the IoT Hub redirection address
        if redirect.address:
            # The regex searches for the Event Hub compatible name in the redirection address. The name is nested in
            # between the port and 'ConsumerGroups'.
            # (ex. "...servicebus.windows.net:12345/<Event Hub name>/ConsumerGroups/...").
            # The regex matches string ':<digits>/', then any characters, then the string '/ConsumerGroups'.
            iot_hub_name = (
                cast(re.Match, re.search(":\d+\/.*/ConsumerGroups", str(redirect.address))).group(0).split("/")[1]
            )
        return "Endpoint=sb://{}/;SharedAccessKeyName={};SharedAccessKey={};EntityPath={}".format(
            fully_qualified_name, shared_access_key_name, shared_access_key, iot_hub_name
        )
    except Exception as exp:
        raise ValueError(
            "{} is an invalid IoT Hub connection string. The underlying exception is {}".format(
                iothub_conn_str,
                exp,
            )
        )


async def receive_events_from_iothub(iothub_conn_str):
    """Convert the iot hub connection string to the built-in eventhub connection string
    and receive events from the eventhub
    """
    eventhub_conn_str = convert_iothub_to_eventhub_conn_str(iothub_conn_str)
    consumer_client = EventHubConsumerClient.from_connection_string(eventhub_conn_str, consumer_group="$Default")

    async def on_event_batch(partition_context, events):
        # put your code here
        print("received {} events from partition {}".format(len(events), partition_context.partition_id))

    async with consumer_client:
        await consumer_client.receive_batch(
            on_event_batch, starting_position=-1  # "-1" is from the beginning of the partition.
        )


if __name__ == "__main__":
    iothub_conn_str = os.environ["IOTHUB_CONN_STR"]
    asyncio.run(receive_events_from_iothub(iothub_conn_str))
